#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>

from requests import exceptions as req_exceptions

from .tools import KEYISSET


class ConnectGenericException(Exception):
    """
    Generic class handling Exceptions
    """

    def __init__(self, msg, code, details):
        """

        :param msg:
        :param code:
        :param details:
        """
        super().__init__(msg, code, details)
        self.code = code
        self.details = details


class GenericNotFound(ConnectGenericException):
    """
    Generic option for 404 return code
    """

    def __init__(self, code, details):
        super().__init__(details.get("detail", "Resource not found"), code, details)


class GenericConflict(ConnectGenericException):
    """
    Generic option for 409 return code
    """

    def __init__(self, code, details):
        super().__init__(details.get("detail", "Resources conflict"), code, details)


class GenericUnauthorized(ConnectGenericException):
    """
    Generic option for 401 return code
    """

    def __init__(self, code, details):
        super().__init__(details.get("detail", "Access unauthorized"), code, details)


class GenericForbidden(ConnectGenericException):
    """
    Generic exception for a 403
    """

    def __init__(self, code, details):
        super().__init__(details.get("detail", "403 Forbidden"), code, details)


class ConnectApiException(ConnectGenericException):
    """
    Top class for DatabaseUser exceptions
    """

    def __init__(self, code, details):
        if code == 409:
            raise GenericConflict(code, details)
        elif code == 404:
            raise GenericNotFound(code, details)
        elif code == 401:
            raise GenericUnauthorized(code, details)
        elif code == 403:
            raise GenericForbidden(code, details)
        super().__init__("Something was wrong with the client request.", code, details)


def evaluate_api_return(function):
    """
    Decorator to evaluate the requests payload returned
    """

    def wrapped_answer(*args, **kwargs):
        """
        Decorator wrapper
        """
        try:
            payload = function(*args, **kwargs)
            if payload.status_code not in [200, 201, 202, 204] and not KEYISSET(
                "ignore_failure", kwargs
            ):
                details = payload.json()
                raise ConnectApiException(payload.status_code, details)

            elif KEYISSET("ignore_failure", kwargs):
                return payload
            return payload
        except req_exceptions.RequestException as error:
            print(error)
            raise

    return wrapped_answer



# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2022 John Mille <john@compose-x.io>

"""Main module."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from requests import Response

import re

from requests import delete, get, post, put
from requests.auth import HTTPBasicAuth

from .errors import evaluate_api_return

LOG_LEVELS = ["INFO", "DEBUG", "TRACE", "WARN", "ERROR", "CRITICAL"]


class Task:
    """
    Class to represent a Connector Task
    """

    def __init__(self, connector: Connector, task_id: int, task_config: dict):
        """
        Initializes the Task for a given connector
        """
        self.id = task_id
        self._connector = connector
        self.config = task_config

    def __repr__(self):
        return f"{self._connector.name}.{self.id}"

    @property
    def api(self) -> Api:
        return self._connector.api

    @property
    def connector(self) -> Connector:
        return self._connector

    @property
    def status(self):
        _query = self.api.get(
            f"/connectors/{self._connector.name}/tasks/{self.id}/status"
        )
        return _query

    @property
    def state(self):
        return self.status["state"]

    def is_running(self):
        if self.state == "RUNNING":
            return True
        return False

    def restart(self):
        _query = self.api.post_raw(
            f"/connectors/{self._connector.name}/tasks/{self.id}/restart"
        )


class Connector:
    """
    Class to represent a Connector.

    Configuration & Tasks are retrieved live from the connect cluster, never cached, to ensure there is no
    conflict or out of date settings.
    """

    def __init__(self, cluster: Cluster, name: str):
        self._cluster = cluster
        self.name = name

    @property
    def api(self) -> Api:
        return self._cluster.api

    @property
    def cluster(self) -> Cluster:
        return self._cluster

    @property
    def connector_class(self) -> str:
        return self.config["connector.class"]

    def __repr__(self):
        return self.name

    def exists(self) -> bool:
        req = self.api.get_raw(f"/connectors/{self.name}/")
        if req.status_code == 404:
            return False
        return True

    def restart(self) -> None:
        self.api.post(f"/connectors/{self.name}/restart")

    def pause(self) -> None:
        self.api.put_raw(f"/connectors/{self.name}/pause")

    def resume(self) -> None:
        self.api.put_raw(f"/connectors/{self.name}/resume")

    def restart_all_tasks(self) -> None:
        for _task in self.tasks:
            _task.restart()

    def delete(self) -> None:
        self.api.delete_raw(f"/connectors/{self.name}")

    def cycle_connector(self) -> None:
        self.pause()
        self.restart_all_tasks()
        self.resume()

    @property
    def status(self) -> dict:
        return self.api.get(f"/connectors/{self.name}/status")

    @property
    def state(self):
        return self.status["connector"]["state"]

    @property
    def config(self):
        _config = self.api.get(f"/connectors/{self.name}")
        return _config["config"]

    @config.setter
    def config(self, config: dict) -> None:
        if not isinstance(config, dict):
            raise TypeError(
                self.name,
                "connect configuration must be a dictionary/mapping. Got",
                type(config),
            )
        _req = self.api.put_raw(f"/connectors/{self.name}/config", json=config)
        if not (199 < _req.status_code < 300):
            print(_req.text)

    @property
    def tasks(self):
        _tasks = []
        _connector_tasks = self.api.get(f"/connectors/{self.name}/tasks")
        for _task in _connector_tasks:
            _tasks.append(
                Task(
                    self,
                    task_id=int(_task["id"]["task"]),
                    task_config=_task["config"],
                )
            )
        return _tasks


class Cluster:

    """
    Class to represent the cluster at the top level.

    Configurations & Connectors are retrieved "live" from the API, not stored in memory,
    to avoid conflicts/out of date settings.
    """

    def __init__(self, api: Api):
        self._api = api

    def get(self):
        return self._api.get("/")

    @property
    def api(self) -> Api:
        return self._api

    @property
    def version(self):
        return self.get()["version"]

    @property
    def kafka_cluster(self) -> str:
        return self.get()["kafka_cluster_id"]

    @property
    def connectors(self) -> dict:
        _connectors = self._api.get("/connectors")
        _cluster_connectors: dict = {}
        for connector in _connectors:
            _cluster_connectors[connector] = Connector(self, connector)
        return _cluster_connectors

    @property
    def loggers(self) -> dict:
        return self._api.get("/admin/loggers")

    @property
    def root_logger(self):
        return self.loggers["root"]

    @root_logger.setter
    def root_logger(self, log_level: str) -> None:
        if log_level not in LOG_LEVELS:
            raise ValueError(log_level, "not valid. Must be one of", LOG_LEVELS)
        self._api.put("/admin/loggers/root", json={"level": log_level})

    def set_logger_log_level(self, logger_name: str, log_level: str) -> dict:
        if log_level not in LOG_LEVELS:
            raise ValueError(log_level, "not valid. Must be one of", LOG_LEVELS)
        if logger_name not in self.loggers.keys():
            raise ValueError(
                logger_name,
                "Logger not found on the cluster. Valid loggers",
                list(self.loggers.keys()),
            )
        return self._api.put(f"/admin/loggers/{logger_name}", json={"level": log_level})

    def __repr__(self):
        return f"{self._api.url} - {self.version} - {self.kafka_cluster}"


class Api:
    """
    API Calls handler. Used by the Connect cluster to wrap connect API calls.
    """

    def __init__(
        self,
        hostname,
        port=None,
        url=None,
        protocol=None,
        ignore_ssl_errors=False,
        username=None,
        password=None,
    ):
        """

        :param str hostname:
        :param int port:
        :param str protocol:
        :param bool ignore_ssl_errors:
        :param str username:
        :param str password:
        """
        self.hostname = hostname
        self.protocol = protocol if protocol else "http"
        self.verify_ssl = not ignore_ssl_errors
        self.port = port if port else 8083
        self.username = username
        self.password = password

        if self.protocol not in ["http", "https"]:
            raise ValueError("protocol must be one of", ["http", "https"])
        if (self.port < 0) or (self.port > (2**16)):
            raise ValueError(
                f"Port {self.port} is not valid. Must be between 0 and {((2 ** 16) - 1)}"
            )
        if self.username and not self.password or self.password and not self.username:
            raise ValueError("You must specify both username and password")
        if self.verify_ssl is True and self.protocol == "http":
            print("No SSL needed for HTTP without TLS. Disabling")
            self.verify_ssl = False
        if url:
            self.url = url
            if not re.match(r"(http://|https://)", self.url):
                print(f"URL Does not contain a protocol. Using default {self.protocol}")
                self.url = f"{self.protocol}://{self.url}"
            print("URL Defined from parameter. Skipping hostname:port parameters")
        elif (self.port == 80 and protocol == "http") or (
            self.port == 443 and self.protocol == "https"
        ):
            self.url = f"{self.protocol}://{self.hostname}"
        else:
            self.url = f"{self.protocol}://{self.hostname}:{self.port}"

        self.auth = (
            HTTPBasicAuth(self.username, self.password)
            if self.username and self.password
            else None
        )

        self.headers = {
            "Content-type": "application/json",
            "Accept": "application/json",
        }

    def __repr__(self):
        return self.url

    @evaluate_api_return
    def get_raw(self, query_path, **kwargs) -> Response:
        if not query_path.startswith(r"/"):
            query_path = f"/{query_path}"
        url = f"{self.url}{query_path}"
        req = get(
            url, auth=self.auth, headers=self.headers, verify=self.verify_ssl, **kwargs
        )
        return req

    def get(self, query_path):
        req = self.get_raw(query_path)
        return req.json()

    @evaluate_api_return
    def post_raw(self, query_path, **kwargs) -> Response:
        if not query_path.startswith(r"/"):
            query_path = f"/{query_path}"
        url = f"{self.url}{query_path}"
        req = post(
            url,
            auth=self.auth,
            headers=self.headers,
            verify=self.verify_ssl,
            **kwargs,
        )
        return req

    def post(self, query_path, **kwargs):
        req = self.post_raw(query_path, **kwargs)
        return req.json()

    @evaluate_api_return
    def put_raw(self, query_path, **kwargs) -> Response:
        if not query_path.startswith(r"/"):
            query_path = f"/{query_path}"
        url = f"{self.url}{query_path}"
        req = put(
            url,
            auth=self.auth,
            headers=self.headers,
            verify=self.verify_ssl,
            **kwargs,
        )
        return req

    def put(self, query_path, **kwargs):
        req = self.put_raw(query_path, **kwargs)
        return req.json()

    @evaluate_api_return
    def delete_raw(self, query_path, **kwargs) -> Response:
        if not query_path.startswith(r"/"):
            query_path = f"/{query_path}"
        url = f"{self.url}{query_path}"
        req = delete(
            url, auth=self.auth, headers=self.headers, verify=self.verify_ssl, **kwargs
        )
        return req

    def delete(self, query_path):
        req = self.delete_raw(query_path)
        return req.json()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kafka_connect_api']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.26.0,<3.0.0']

extras_require = \
{':extra == "awslambda"': ['jsonschema>=4.3.0,<5.0.0']}

setup_kwargs = {
    'name': 'kafka-connect-api',
    'version': '0.5.2',
    'description': 'Apache Kafka Connect client',
    'long_description': '========================================\nApache Kafka Connect API Python client\n========================================\n\n|PYPI_VERSION| |PYPI_LICENSE|\n\n|CODE_STYLE| |TDD| |BDD|\n\n|DOCS_BUILD|\n\nDocumentation: https://kafka-connect-api.readthedocs.io.\n\n\nInstall\n========\n\n.. code-block::\n\n    pip install kafka-connect-api\n\nUsage\n======\n\n.. code-block:: python\n\n    from kafka_connect_api.kafka_connect_api import Api, Cluster\n\n    api = Api(connect.cluster, port=8083)\n    cluster = Cluster(api)\n    print(cluster.connectors)\n\n\nFeatures\n==========\n\nAllows you to interact with the Kafka Connect API (`API Reference`_) in a simple way.\n\n* Connection to cluster (supports Basic Auth)\n* List all connectors in the cluster\n* Describe the connector configuration\n* Update the connector configuration\n* Pause / Resume Connector\n* Restart all tasks for the connector\n\nSome pre-made functions can help with operational activities.\nSee `kafka_connect_api.aws_lambdas.py`\n\n.. _API Reference: https://docs.confluent.io/platform/current/connect/references/restapi.html\n\n.. |DOCS_BUILD| image:: https://readthedocs.org/projects/kafka-connect-api/badge/?version=latest\n        :target: https://kafka-connect-api.readthedocs.io/en/latest/\n        :alt: Documentation Status\n\n.. |PYPI_VERSION| image:: https://img.shields.io/pypi/v/kafka-connect-api.svg\n        :target: https://pypi.python.org/pypi/kafka_connect_api\n\n.. |PYPI_LICENSE| image:: https://img.shields.io/pypi/l/kafka-connect-api\n    :alt: PyPI - License\n    :target: https://github.com/compose-x/kafka-connect-api/blob/master/LICENSE\n\n.. |PYPI_PYVERS| image:: https://img.shields.io/pypi/pyversions/kafka-connect-api\n    :alt: PyPI - Python Version\n    :target: https://pypi.python.org/pypi/kafka-connect-api\n\n.. |PYPI_WHEEL| image:: https://img.shields.io/pypi/wheel/kafka-connect-api\n    :alt: PyPI - Wheel\n    :target: https://pypi.python.org/pypi/kafka-connect-api\n\n.. |CODE_STYLE| image:: https://img.shields.io/badge/codestyle-black-black\n    :alt: CodeStyle\n    :target: https://pypi.org/project/black/\n\n.. |TDD| image:: https://img.shields.io/badge/tdd-pytest-black\n    :alt: TDD with pytest\n    :target: https://docs.pytest.org/en/latest/contents.html\n\n.. |BDD| image:: https://img.shields.io/badge/bdd-behave-black\n    :alt: BDD with Behave\n    :target: https://behave.readthedocs.io/en/latest/\n\n.. |QUALITY| image:: https://sonarcloud.io/api/project_badges/measure?project=compose-x_kafka-connect-api&metric=alert_status\n    :alt: Code scan with SonarCloud\n    :target: https://sonarcloud.io/dashboard?id=compose-x_kafka-connect-api\n\n.. |PY_DLS| image:: https://img.shields.io/pypi/dm/kafka-connect-api\n    :target: https://pypi.org/project/kafka-connect-api/\n',
    'author': 'John Preston',
    'author_email': 'john@compose-x.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
